/*
Copyright (c) 2009-2010 Mikko Mononen memon@inside.org
recast4j copyright (c) 2015-2019 Piotr Piastucki piotr@jtilia.org
DotRecast Copyright (c) 2023-2024 Choi Ikpil ikpil@naver.com

This software is provided 'as-is', without any express or implied
warranty.  In no event will the authors be held liable for any damages
arising from the use of this software.
Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:
1. The origin of this software must not be misrepresented; you must not
 claim that you wrote the original software. If you use this software
 in a product, an acknowledgment in the product documentation would be
 appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
 misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using DotRecast.Core.Buffers;

namespace DotRecast.Detour
{
    public class DtNodePool
    {
        private readonly Dictionary<long, List<DtNode>> m_map;

        private int m_usedNodesCount;
        private List<DtNode[]> m_buckets;
        private readonly int m_initialBufferCapacityBase;
        private readonly RcObjectPool<List<DtNode>> m_listPool;

        public DtNodePool(int initialBufferCapacityBase = 6) // initial size 64
        {
            m_map = new Dictionary<long, List<DtNode>>();
            m_listPool = new RcObjectPool<List<DtNode>>(() => new List<DtNode>());
            m_buckets = new List<DtNode[]>();
            m_initialBufferCapacityBase = initialBufferCapacityBase;
        }

        private void AddNewBucket()
        {
            var bucketIndex = m_buckets.Count;
            var bucket = new DtNode[1 << (bucketIndex + m_initialBufferCapacityBase)];
            m_buckets.Add(bucket);
            FillBucket(bucketIndex);
        }

        private void FillBucket(int bucketIndex)
        {
            var bucket = m_buckets[bucketIndex];
            var startIndex = GetBucketStartIndex(bucketIndex);
            for (int i = 0; i < bucket.Length; i++)
            {
                bucket[i] = new DtNode(startIndex + i);
            }
        }

        private int GetBucketStartIndex(int bucketIndex)
        {
            return ((1 << (bucketIndex + m_initialBufferCapacityBase)) - 1) ^ ((1 << m_initialBufferCapacityBase) - 1);
        }

        private DtNode[] EnsureBucket(int bucketIndex)
        {
            if (m_buckets.Count == bucketIndex)
                AddNewBucket();
            else if (m_buckets.Count < bucketIndex)
                throw new Exception();
            return m_buckets[bucketIndex];
        }

        private int GetBucketIndexByElementIndex(int elementIndex)
        {
            return DtUtils.Ilog2((elementIndex >> m_initialBufferCapacityBase) + 1);
        }

        public void Clear()
        {
            foreach (var pair in m_map)
                m_listPool.Return(pair.Value);

            m_map.Clear();
            m_usedNodesCount = 0;
        }

        public int GetNodeCount()
        {
            return m_usedNodesCount;
        }

        public int FindNodes(long id, out List<DtNode> nodes)
        {
            var hasNode = m_map.TryGetValue(id, out nodes);
            if (hasNode)
            {
                return nodes.Count;
            }

            return 0;
        }

        public DtNode FindNode(long id)
        {
            m_map.TryGetValue(id, out var nodes);
            if (nodes != null && 0 != nodes.Count)
            {
                return nodes[0];
            }

            return null;
        }

        public DtNode GetNode(long id, int state)
        {
            m_map.TryGetValue(id, out var nodes);
            if (nodes != null)
            {
                foreach (DtNode node in nodes)
                {
                    if (node.state == state)
                    {
                        return node;
                    }
                }
            }
            else
            {
                nodes = m_listPool.Get();
                nodes.Clear();
                m_map.Add(id, nodes);
            }

            return Create(id, state, nodes);
        }

        private DtNode Create(long id, int state, List<DtNode> nodes)
        {
            int i = m_usedNodesCount++;
            int bucketIndex = GetBucketIndexByElementIndex(i);
            int bucketStartIndex = GetBucketStartIndex(bucketIndex);
            var node = EnsureBucket(bucketIndex)[i - bucketStartIndex];
            node.pidx = 0;
            node.cost = 0;
            node.total = 0;
            node.id = id;
            node.state = state;
            node.flags = 0;
            node.shortcut = null;

            nodes.Add(node);
            return node;
        }

        public int GetNodeIdx(DtNode node)
        {
            return node != null
                ? node.ptr + 1
                : 0;
        }

        public DtNode GetNodeAtIdx(int idx)
        {
            if (idx == 0)
                return null;
            
            int bucketIndex = GetBucketIndexByElementIndex(idx - 1);
            if (m_buckets.Count <= bucketIndex)
                throw new ArgumentOutOfRangeException();

            int bucketStartIndex = GetBucketStartIndex(bucketIndex);
            var node = EnsureBucket(bucketIndex)[idx - bucketStartIndex - 1];
            return node;
        }

        public DtNode GetNode(long refs)
        {
            return GetNode(refs, 0);
        }

        public IEnumerable<DtNode> AsEnumerable()
        {
            return m_buckets.SelectMany(x => x);
        }
    }
}